//
// Created by Tom on 02/08/2023.
//

#ifndef TYPON_SLICE_HPP
#define TYPON_SLICE_HPP

#include <optional>
#include <utility>
#include <Python.h>
#include <stdint.h>

struct PySlice {
  PySlice() = default;
  PySlice(const PySlice &) = default;
  PySlice(PySlice &&) = default;
  PySlice &operator=(const PySlice &) = default;
  PySlice &operator=(PySlice &&) = default;

  PySlice(std::optional<ssize_t> start, std::optional<ssize_t> stop, std::optional<ssize_t> step) : start(start), stop(stop), step(step) {
    if (step == 0) {
      throw std::runtime_error("slice step cannot be zero");
    }
  }

  std::optional<ssize_t> start = 0;
  std::optional<ssize_t> stop = 0;
  std::optional<ssize_t> step = 1;

  struct UnpackedSlice {
    ssize_t start, stop, step;
  };

  std::pair<ssize_t, UnpackedSlice> adjust_indices(ssize_t seq_length) {
    UnpackedSlice res;

    if (this->step == std::nullopt) {
      res.step = 1;
    } else {
      res.step = this->step.value();
    }

    if (this->start == std::nullopt) {
      res.start = res.step < 0 ? PY_SSIZE_T_MAX : 0;
    }

    if (this->stop == std::nullopt) {
      res.stop = res.step < 0 ? PY_SSIZE_T_MIN : PY_SSIZE_T_MAX;
    }

    auto len = PySlice_AdjustIndices(seq_length, &res.start, &res.stop, res.step);

    return {len, res};
  }
};

#endif // TYPON_SLICE_HPP
