//
// Created by Tom on 08/03/2023.
//

#ifndef TYPON_STR_HPP
#define TYPON_STR_HPP

#include <sstream>
#include <string>

using namespace std::literals;

#include "bytes.hpp"
#include "print.hpp"
#include "slice.hpp"
// #include <format>
#include <fmt/format.h>

class PyStr : public std::string {
public:
  PyStr() : std::string() {}
  PyStr(const std::string &s) : std::string(s) {}
  PyStr(std::string &&s) : std::string(std::move(s)) {}
  constexpr PyStr(const char *s, size_t count) : std::string(s, count) {}
  constexpr PyStr(size_t count, char ch) : std::string(count, ch) {}
  template<typename... Args>
  PyStr(Args&&... args) : std::string(std::forward<Args>(args)...) {}

  template <class InputIterator>
  PyStr(InputIterator first, InputIterator last) : std::string(first, last) {}


  METHOD(PyBytes, encode, (Self self, const std::string &encoding = "utf-8"), {
        return PyBytes(self.begin(), self.end());
          })

  METHOD_GEN((typename... T), PyStr, format, (Self self, T &&...args), {
        return PyStr(fmt::format(fmt::runtime(self), std::forward<T>(args)...));
          })

  METHOD(bool, startswith, (Self self, const std::string &s), {
        return self.starts_with(s);
          })

  METHOD(int, find, (Self self, const std::string &s), {
        auto pos = self.std::string::find(s);
        return pos == std::string::npos ? -1 : pos;
          })

  PyStr operator[](PySlice slice) const {
    auto [len, new_slice] = slice.adjust_indices(this->size());

    PyStr result;
    result.reserve(len);

    if (new_slice.start < new_slice.stop) {
      if (new_slice.step > 0) {
        for (auto i = new_slice.start; i < new_slice.stop;
             i += new_slice.step) {
          result.push_back(this->char_at(i));
        }
      }
    } else {
      if (new_slice.step < 0) {
        for (auto i = new_slice.start; i > new_slice.stop;
             i += new_slice.step) {
          result.push_back(this->char_at(i));
        }
      }
    }

    return result;
  }

  PyStr operator[](ssize_t index) const {
    if (index < 0) {
      index += this->size();
    }

    return PyStr(1, std::string::operator[](index));
  }

  char char_at(ssize_t index) const {
    if (index < 0) {
      index += this->size();
    }

    return this->begin()[index];
  }

  operator int() const {
    return std::stoi(*this);
  }
};

inline constexpr PyStr operator""_ps(const char *s, size_t len) noexcept {
  return PyStr(s, len);
}

template<typename Self>
PyStr PyBytes::decode_s::operator()(Self self, const std::string &encoding) const {
  return PyStr(self.begin(), self.end());
}

template <typename T> PyStr str(const T &x) {
  std::stringstream s;
  print_to(x, s);
  return s.str();
}

template <typename T> PyStr repr(const T &x) {
  std::stringstream s;
  repr_to(x, s);
  return s.str();
}

template <> struct std::hash<PyStr> {
  std::size_t operator()(const PyStr &s) const noexcept {
    return std::hash<std::string>()(s);
  }
};

#endif // TYPON_STR_HPP
