//
// Created by Tom on 13/06/2023.
//

#ifndef TYPON_SOCKET_HPP
#define TYPON_SOCKET_HPP

#include "builtins.hpp"
#include <netinet/in.h>
#include <tuple>

namespace py_socket {
struct socket_t {
#undef SOCK_STREAM
#undef AF_INET6
#undef SOL_SOCKET
#undef SO_REUSEADDR
#undef AF_UNIX
  static constexpr int SOCK_STREAM = 1;
  static constexpr int AF_INET6 = 10;
  static constexpr int SOL_SOCKET = 1;
  static constexpr int SO_REUSEADDR = 2;
  static constexpr int AF_UNIX = 1;

  struct socket_s {
    struct type {
      METHOD(typon::Task<std::tuple<PyObj<type> COMMA() std::string>>, accept, (Self self), {
        int connfd = co_await typon::io::accept(self->fd, NULL, NULL);
        if (connfd < 0) {
          system_error(-connfd, "accept()");
        }
        co_return std::make_tuple(pyobj<type>(connfd), std::string("")); // TODO
      })

      METHOD(typon::Task<void>, close, (Self self),
             { co_await typon::io::close(self->fd); })

      METHOD(void, listen, (Self self, int backlog), {
        if (::listen(self->fd, backlog) < 0) {
          dotp(self, close)();
          system_error(errno, "listen()");
        }
      })

      METHOD(void, setsockopt, (Self self, int level, int optname, int optval), {
        if (::setsockopt(self->fd, level, optname, &optval, sizeof(int)) < 0) {
          system_error(errno, "setsockopt()");
        }
      })

      METHOD(void, bind, (Self self, std::tuple<std::string COMMA() int> address), {
        auto [host, port] = address;
        sockaddr_in6 addr;
        std::memset(&addr, 0, sizeof(addr));
        addr.sin6_family = AF_INET6;
        addr.sin6_port = htons(port);
        addr.sin6_addr = in6addr_any;
        if (::bind(self->fd, (const sockaddr *)&addr, sizeof(addr)) < 0) {
          system_error(errno, "bind()");
        }
      })

      METHOD(typon::Task<PyBytes>, recv, (Self self, int bufsize), {
        PyBytes buf(bufsize, '\0');
        co_await typon::io::recv(self->fd, buf.data(), buf.size(), 0);
        co_return std::move(buf);
      })

      METHOD(typon::Task<void>, send, (Self self, PyBytes data), {
        if (int sbytes = co_await typon::io::send(self->fd, data, 0);
            sbytes < 0) {
          co_await dotp(self, close)();
          system_error(-sbytes, "send()");
        }
      })

      type(int fd = -1) : fd(fd) {}

      type(const type &other)
          : fd(other.fd) {}

      int fd;
    };

    auto operator()(int family, int type_) {
      if (int fd = ::socket(family, type_, 0); fd >= 0) {
        return pyobj<type>(fd);
      } else {
        system_error(errno, "socket()");
      }
    }
  } socket;

  FUNCTION(auto, getaddrinfo, (std::string host, int port, int family=0, int type_=0, int proto=0, int flags=0), {
    /*addrinfo hints;
    std::memset(&hints, 0, sizeof(hints));
    hints.ai_family = family;
    hints.ai_socktype = type_;
    hints.ai_protocol = proto;
    hints.ai_flags = flags;
    addrinfo *res;
    if (int err = ::getaddrinfo(host.c_str(), port.c_str(), &hints, &res); err != 0) {
      system_error(err, "getaddrinfo()");
    }
    return res;*/
                                                                                                                                     return 123;
  })
} all;

auto &get_all() { return all; }
} // namespace py_socket

namespace typon {
using PySocket = PyObj<py_socket::socket_t::socket_s>;
}

#endif // TYPON_SOCKET_HPP
